package arzumify.presence.presences;

import arzumify.presence.maths.Vec2i;
import arzumify.presence.maths.Vec3i;

import java.util.*;

public class IntegerPresence implements Presence<IntegerPresence> {
    public IntegerPresence(short range, Vec3i... basePositions) {
        this.basePositions.addAll(List.of(basePositions));
        this.range = range;
        this.baseChunkPresence = getChunkPresence(this.basePositions);
        this.collisionChunkPresence = getChunkPresence(ComputeCollisionShape(range, this.basePositions));
    }

    private final HashSet<Vec3i> basePositions = new HashSet<>();
    public final HashSet<Vec2i> baseChunkPresence;
    public final HashSet<Vec2i> collisionChunkPresence;
    private final short range;

    // Calculates the chunk presence for the given positions
    public HashSet<Vec2i> getChunkPresence(HashSet<Vec3i> positions) {
        var chunkPresence = new HashSet<Vec2i>();
        for (Vec3i basePosition : positions) {
            chunkPresence.add(new Vec2i(basePosition.x() >> 4, basePosition.z() >> 4));
        }
        return chunkPresence;
    }

    @Override
    public boolean Matches(Presence<IntegerPresence> presence) {
        return presence instanceof IntegerPresence && matches((IntegerPresence) presence);
    }

    private boolean matches(IntegerPresence other) {
        // Take the other's base position and calculate the manhattan distance to each of this presence's positions
        for (Vec3i otherBasePosition : other.basePositions) {
            for (Vec3i presence : this.basePositions) {
                if (ManhattanDistance(otherBasePosition, presence) <= this.range) {
                    return true;
                } else {
                    // Print the manhattan distance
                    System.out.println("Manhattan distance between " + otherBasePosition + " and " + presence + " is " + ManhattanDistance(otherBasePosition, presence));
                }
            }
        }
        return false;
    }

    public int ManhattanDistance(Vec3i a, Vec3i b) {
        return Math.abs(a.x() - b.x()) + Math.abs(a.y() - b.y()) + Math.abs(a.z() - b.z());
    }

    private static final Map<Short, HashSet<Vec3i>> cachedCollisionShapes = new HashMap<>();

    private static HashSet<Vec3i> ComputeCollisionShape(short range, Set<Vec3i> basePositions) {
        if (!cachedCollisionShapes.containsKey(range)) {
            var possiblePresences = ComputePossiblePresences(range);
            cachedCollisionShapes.put(range, possiblePresences);
        }
        var collisionShape = new HashSet<Vec3i>();
        for (Vec3i basePosition : basePositions) {
            for (Vec3i possiblePresence : cachedCollisionShapes.get(range)) {
                collisionShape.add(basePosition.add(possiblePresence));
            }
        }
        return collisionShape;
    }

    public static HashSet<Vec3i> ComputePossiblePresences(short range) {
        var presences = new HashSet<Vec3i>();
        for (int dx = -range; dx <= range; dx++) {
            for (int dy = -range; dy <= range; dy++) {
                int dzMax = range - (Math.abs(dx) + Math.abs(dy));
                if (dzMax >= 0) {
                    for (int dz = -dzMax; dz <= dzMax; dz++) {
                        presences.add(new Vec3i(dx, dy, dz));
                    }
                }
            }
        }
        return presences;
    }

    @Override
    public HashSet<Vec2i> GetChunks() {
        return baseChunkPresence;
    }

    @Override
    public HashSet<Vec2i> GetCollisionChunks() {
        return collisionChunkPresence;
    }
}
