import arzumify.presence.matchmaker.Channel;
import arzumify.presence.matchmaker.MatchMaker;
import arzumify.presence.maths.Vec3f;
import arzumify.presence.presences.FloatingPresence;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;

import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import static arzumify.presence.presences.FloatingPresence.ComputePossiblePresences;
import static org.junit.jupiter.api.Assertions.*;

@Timeout(1)
public class FloatingPresenceTests {
    private final Channel global = new Channel();
    private final Random rand = new Random();

    @Test
    public void testOffsetCalculations() {
        System.out.println("Testing offset calculations");
        var set = ComputePossiblePresences((short) 1);
        System.out.println(set);
        assertTrue(set.containsAll(Set.of(
                new Vec3f(1, 0, 0),
                new Vec3f(0, 1, 0),
                new Vec3f(0, 0, 1),
                new Vec3f(-1, 0, 0),
                new Vec3f(0, -1, 0),
                new Vec3f(0, 0, -1)
        )));
    }

    @Test
    public void testPresenceCalculations() {
        System.out.println("Testing presence calculations");
        var presence = new FloatingPresence((short) 1, new Vec3f(0, 0, 0));
        var presence2 = new FloatingPresence((short) 1, new Vec3f(1, 0, 0));
        assertTrue(presence.Matches(presence2));
    }

    @Nested
    class TestMatchMaker {
        private final MatchMaker<FloatingPresence, ExampleFloatingPresence> matchMaker = new MatchMaker<>();
        private final ExampleFloatingPresence presence = new ExampleFloatingPresence(new FloatingPresence((short) 1, new Vec3f(0, 0, 0)));
        private final ExampleFloatingPresence presence2 = new ExampleFloatingPresence(new FloatingPresence((short) 1, new Vec3f(1, 0, 0)));

        @Test
        public void testMatchMaker() {
            System.out.println("Testing match maker");
            matchMaker.Add(presence, global);
            matchMaker.Add(presence2, global);
            var matches = matchMaker.Search(presence, global);
            assertEquals(1, matches.size());
            System.out.println(matches);
            assertTrue(matches.contains(presence2));
        }
    }

    @Nested
    class TestMatchMakerLongRange {
        private final MatchMaker<FloatingPresence, ExampleFloatingPresence> matchMaker = new MatchMaker<>();
        private final ExampleFloatingPresence presence = new ExampleFloatingPresence(new FloatingPresence((short) 64, new Vec3f(0, 0, 0)));
        private final List<Vec3f> validPresences = ComputePossiblePresences((short) 64).stream().toList();
        private final ExampleFloatingPresence presence2 = new ExampleFloatingPresence(new FloatingPresence((short) 64, validPresences.get(rand.nextInt(validPresences.size() - 1))));

        @Test
        public void testMatchMakerLongRange() {
            System.out.println("Testing match maker long range");
            matchMaker.Add(presence, global);
            matchMaker.Add(presence2, global);
            var matches = matchMaker.Search(presence, global);
            assertEquals(1, matches.size());
            assertTrue(matches.contains(presence2));
        }
    }

    @Nested
    class TestMatchMakerManyBasePositions {
        private final MatchMaker<FloatingPresence, ExampleFloatingPresence> matchMaker = new MatchMaker<>();
        private final ExampleFloatingPresence presence = new ExampleFloatingPresence(new FloatingPresence((short) 1, new Vec3f(0, 0, 0), new Vec3f(1, 0, 0)));
        private final ExampleFloatingPresence presence2 = new ExampleFloatingPresence(new FloatingPresence((short) 1, new Vec3f(2, 0, 0), new Vec3f(3, 0, 0)));

        @Test
        public void testMatchMakerManyBasePositions() {
            System.out.println("Testing match maker many base positions");
            matchMaker.Add(presence, global);
            matchMaker.Add(presence2, global);
            var matches = matchMaker.Search(presence, global);
            assertEquals(1, matches.size());
            assertTrue(matches.contains(presence2));
        }
    }

    @Nested
    class TestMatchMakerManyBasePositionsLongRange {
        private final MatchMaker<FloatingPresence, ExampleFloatingPresence> matchMaker = new MatchMaker<>();
        private final ExampleFloatingPresence presence = new ExampleFloatingPresence(new FloatingPresence((short) 64, new Vec3f(0, 0, 0), new Vec3f(1, 0, 0)));
        private final ExampleFloatingPresence presence2 = new ExampleFloatingPresence(new FloatingPresence((short) 64, new Vec3f(65, 0, 0), new Vec3f(66, 0, 0)));

        @Test
        public void testMatchMakerManyBasePositionsLongRange() {
            System.out.println("Testing match maker many base positions long range");
            matchMaker.Add(presence, global);
            matchMaker.Add(presence2, global);
            var matches = matchMaker.Search(presence, global);
            assertEquals(1, matches.size());
            assertTrue(matches.contains(presence2));
        }
    }

    @Nested
    class StressTestRanges {
        private final MatchMaker<FloatingPresence, ExampleFloatingPresence> matchMaker = new MatchMaker<>();
        private final Set<ExampleFloatingPresence> presenceSet = new HashSet<>();
        private final ExampleFloatingPresence originalPresence = new ExampleFloatingPresence(new FloatingPresence((short) 32, new Vec3f(0, 0, 0)));

        public StressTestRanges() {
            for (int i = 0; i < 50; i++) {
                List<Vec3f> validPresences = ComputePossiblePresences((short) 32).stream().toList();
                var presence = new ExampleFloatingPresence(new FloatingPresence((short) 32, validPresences.get(rand.nextInt(validPresences.size() - 1))));
                presenceSet.add(presence);
            }
        }

        @Test
        @Timeout(10)
        public void stressTestRange() {
            for (ExampleFloatingPresence presence : presenceSet) {
                matchMaker.Add(presence, global);
            }
            var matches = matchMaker.Search(originalPresence, global);
            assertEquals(presenceSet.size(), matches.size());
            assertFalse(matches.contains(originalPresence));
        }
    }

    @Nested
    class StressTestBasePositions {
        private final MatchMaker<FloatingPresence, ExampleFloatingPresence> matchMaker = new MatchMaker<>();
        private final Set<ExampleFloatingPresence> presenceSet = new HashSet<>();
        private final ExampleFloatingPresence originalPresence = new ExampleFloatingPresence(new FloatingPresence((short) 1, GenerateVec3fsFromArea(new Vec3f(0, 0, 0), new Vec3f(0, 0, 50))));

        public StressTestBasePositions() {
            for (int i = 0; i < 50; i++) {
                var presence = new ExampleFloatingPresence(new FloatingPresence((short) 1, new Vec3f(0, 1, i)));
                presenceSet.add(presence);
            }
        }

        @Test
        @Timeout(10)
        public void stressTestBasePositions() {
            for (ExampleFloatingPresence presence : presenceSet) {
                matchMaker.Add(presence, global);
            }
            var matches = matchMaker.Search(originalPresence, global);
            assertEquals(presenceSet.size(), matches.size());
            assertFalse(matches.contains(originalPresence));
        }
    }

    private Vec3f[] GenerateVec3fsFromArea(Vec3f corner1, Vec3f corner2) {
        var Vec3fs = new HashSet<Vec3f>();

        float minX = Math.min(corner1.x(), corner2.x());
        float maxX = Math.max(corner1.x(), corner2.x());
        float minY = Math.min(corner1.y(), corner2.y());
        float maxY = Math.max(corner1.y(), corner2.y());
        float minZ = Math.min(corner1.z(), corner2.z());
        float maxZ = Math.max(corner1.z(), corner2.z());

        for (float x = minX; x <= maxX; x += (float) 1) {
            for (float y = minY; y <= maxY; y += (float) 1) {
                for (float z = minZ; z <= maxZ; z += (float) 1) {
                    Vec3fs.add(new Vec3f(x, y, z));
                }
            }
        }

        return Vec3fs.toArray(new Vec3f[0]);
    }
}
