import arzumify.presence.matchmaker.Channel;
import arzumify.presence.matchmaker.MatchMaker;
import arzumify.presence.maths.Vec3i;
import arzumify.presence.presences.IntegerPresence;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;

import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import static arzumify.presence.presences.IntegerPresence.ComputePossiblePresences;
import static org.junit.jupiter.api.Assertions.*;

@Timeout(1)
public class IntegerPresenceTests {
    private final Channel global = new Channel();
    private final Random rand = new Random();

    @Test
    public void testOffsetCalculations() {
        System.out.println("Testing offset calculations");
        var set = ComputePossiblePresences((short) 1);
        System.out.println(set);
        assertTrue(set.containsAll(Set.of(
                new Vec3i(1, 0, 0),
                new Vec3i(0, 1, 0),
                new Vec3i(0, 0, 1),
                new Vec3i(-1, 0, 0),
                new Vec3i(0, -1, 0),
                new Vec3i(0, 0, -1)
        )));
    }

    @Test
    public void testPresenceCalculations() {
        System.out.println("Testing presence calculations");
        var presence = new IntegerPresence((short) 1, new Vec3i(0, 0, 0));
        var presence2 = new IntegerPresence((short) 1, new Vec3i(1, 0, 0));
        assertTrue(presence.Matches(presence2));
    }

    @Nested
    class TestMatchMaker {
        private final MatchMaker<IntegerPresence, ExampleIntegerPresence> matchMaker = new MatchMaker<>();
        private final ExampleIntegerPresence presence = new ExampleIntegerPresence(new IntegerPresence((short) 1, new Vec3i(0, 0, 0)));
        private final ExampleIntegerPresence presence2 = new ExampleIntegerPresence(new IntegerPresence((short) 1, new Vec3i(1, 0, 0)));

        @Test
        public void testMatchMaker() {
            System.out.println("Testing match maker");
            matchMaker.Add(presence, global);
            matchMaker.Add(presence2, global);
            var matches = matchMaker.Search(presence, global);
            assertEquals(1, matches.size());
            System.out.println(matches);
            assertTrue(matches.contains(presence2));
        }
    }

    @Nested
    class TestMatchMakerLongRange {
        private final MatchMaker<IntegerPresence, ExampleIntegerPresence> matchMaker = new MatchMaker<>();
        private final ExampleIntegerPresence presence = new ExampleIntegerPresence(new IntegerPresence((short) 64, new Vec3i(0, 0, 0)));
        private final List<Vec3i> validPresences = ComputePossiblePresences((short) 64).stream().toList();
        private final ExampleIntegerPresence presence2 = new ExampleIntegerPresence(new IntegerPresence((short) 64, validPresences.get(rand.nextInt(validPresences.size() - 1))));

        @Test
        public void testMatchMakerLongRange() {
            System.out.println("Testing match maker long range");
            matchMaker.Add(presence, global);
            matchMaker.Add(presence2, global);
            var matches = matchMaker.Search(presence, global);
            assertEquals(1, matches.size());
            assertTrue(matches.contains(presence2));
        }
    }

    @Nested
    class TestMatchMakerManyBasePositions {
        private final MatchMaker<IntegerPresence, ExampleIntegerPresence> matchMaker = new MatchMaker<>();
        private final ExampleIntegerPresence presence = new ExampleIntegerPresence(new IntegerPresence((short) 1, new Vec3i(0, 0, 0), new Vec3i(1, 0, 0)));
        private final ExampleIntegerPresence presence2 = new ExampleIntegerPresence(new IntegerPresence((short) 1, new Vec3i(2, 0, 0), new Vec3i(3, 0, 0)));

        @Test
        public void testMatchMakerManyBasePositions() {
            System.out.println("Testing match maker many base positions");
            matchMaker.Add(presence, global);
            matchMaker.Add(presence2, global);
            var matches = matchMaker.Search(presence, global);
            assertEquals(1, matches.size());
            assertTrue(matches.contains(presence2));
        }
    }

    @Nested
    class TestMatchMakerManyBasePositionsLongRange {
        private final MatchMaker<IntegerPresence, ExampleIntegerPresence> matchMaker = new MatchMaker<>();
        private final ExampleIntegerPresence presence = new ExampleIntegerPresence(new IntegerPresence((short) 64, new Vec3i(0, 0, 0), new Vec3i(1, 0, 0)));
        private final ExampleIntegerPresence presence2 = new ExampleIntegerPresence(new IntegerPresence((short) 64, new Vec3i(65, 0, 0), new Vec3i(66, 0, 0)));

        @Test
        public void testMatchMakerManyBasePositionsLongRange() {
            System.out.println("Testing match maker many base positions long range");
            matchMaker.Add(presence, global);
            matchMaker.Add(presence2, global);
            var matches = matchMaker.Search(presence, global);
            assertEquals(1, matches.size());
            assertTrue(matches.contains(presence2));
        }
    }

    @Nested
    class StressTestRanges {
        private final MatchMaker<IntegerPresence, ExampleIntegerPresence> matchMaker = new MatchMaker<>();
        private final Set<ExampleIntegerPresence> presenceSet = new HashSet<>();
        private final ExampleIntegerPresence originalPresence = new ExampleIntegerPresence(new IntegerPresence((short) 32, new Vec3i(0, 0, 0)));

        public StressTestRanges() {
            for (int i = 0; i < 50; i++) {
                List<Vec3i> validPresences = ComputePossiblePresences((short) 32).stream().toList();
                var presence = new ExampleIntegerPresence(new IntegerPresence((short) 32, validPresences.get(rand.nextInt(validPresences.size() - 1))));
                presenceSet.add(presence);
            }
        }

        @Test
        @Timeout(10)
        public void stressTestRange() {
            for (ExampleIntegerPresence presence : presenceSet) {
                matchMaker.Add(presence, global);
            }
            var matches = matchMaker.Search(originalPresence, global);
            assertEquals(presenceSet.size(), matches.size());
            assertFalse(matches.contains(originalPresence));
        }
    }

    @Nested
    class StressTestBasePositions {
        private final MatchMaker<IntegerPresence, ExampleIntegerPresence> matchMaker = new MatchMaker<>();
        private final Set<ExampleIntegerPresence> presenceSet = new HashSet<>();
        private final ExampleIntegerPresence originalPresence = new ExampleIntegerPresence(new IntegerPresence((short) 1, GenerateVec3isFromArea(new Vec3i(0, 0, 0), new Vec3i(0, 0, 50))));

        public StressTestBasePositions() {
            for (int i = 0; i < 50; i++) {
                var presence = new ExampleIntegerPresence(new IntegerPresence((short) 1, new Vec3i(0, 1, i)));
                presenceSet.add(presence);
            }
        }

        @Test
        @Timeout(10)
        public void stressTestBasePositions() {
            for (ExampleIntegerPresence presence : presenceSet) {
                matchMaker.Add(presence, global);
            }
            var matches = matchMaker.Search(originalPresence, global);
            assertEquals(presenceSet.size(), matches.size());
            assertFalse(matches.contains(originalPresence));
        }
    }

    private Vec3i[] GenerateVec3isFromArea(Vec3i corner1, Vec3i corner2) {
        var vec3is = new HashSet<Vec3i>();

        int minX = Math.min(corner1.x(), corner2.x());
        int maxX = Math.max(corner1.x(), corner2.x());
        int minY = Math.min(corner1.y(), corner2.y());
        int maxY = Math.max(corner1.y(), corner2.y());
        int minZ = Math.min(corner1.z(), corner2.z());
        int maxZ = Math.max(corner1.z(), corner2.z());

        for (int x = minX; x <= maxX; x++) {
            for (int y = minY; y <= maxY; y++) {
                for (int z = minZ; z <= maxZ; z++) {
                    vec3is.add(new Vec3i(x, y, z));
                }
            }
        }

        return vec3is.toArray(new Vec3i[0]);
    }

}
